#!/usr/bin/python

#from gtk import *
import gtk

program_name = "pedit"
program_version = "0.3"

from tktk import ButtonChoiceWindow

# TODO: 
# - fix or work around g_utf8_validate error when inserting non-ascii text
# + keybindings for cut/copy/paste, and maybe more
# * set / change font
# * make "mini" or "embedded" mode for use in pyrpn, etc
# * fix cwd issues
# * implement "File -> New"
# * place cursor at beginning of buffer after loading

class EditWidget(gtk.TextView):
  def __init__(self):
    gtk.TextView.__init__(self)
    self._buffer = self.get_buffer()
    self._font_tag = None

    import time
    self.time = time

    self.linewrap = 1
    self.set_editable(1)
    self.set_cursor_visible(1)
    self.set_wrap_mode(gtk.WRAP_WORD)
    #self.set_line_wrap(self.linewrap)
    #self.text = ""
    self.font = None
    
    # handle cut/copy/paste from X11 clipboard
    self.connect("selection_clear_event", self._selection_clear)
    self.selection_add_target("PRIMARY", "STRING", 1)
    self.selection_add_target("PRIMARY", "COMPOUND_TEXT", 1)
    self.connect("selection_get", self._selection_handle)
    self.connect("selection_received", self._selection_received)

  def set_text(self, text):
    self._buffer.set_text("")
    pos = 0
    # FIXME: g_utf8_validate() errors when loading some texts
    # (problem is semi-worked-around by loading smaller chunks, but
    #  it's still pretty bad to have a chunk missing from the buffer)
    # (GTK should let me know if it has an error!)
    while pos < len(text):
      #print "pos: %s" % (pos)
      t = text[pos:pos+1024]
      self._buffer.insert_at_cursor(t)
      pos += 1024
    #self._buffer.set_text(text)
    iter = self._buffer.get_start_iter()
    self._buffer.place_cursor(iter)

  def get_text(self):
    start, end = self._buffer.get_bounds()
    return self._buffer.get_text(start, end)
 
  def clear(self):
    self._buffer.set_text("")

  def set_font(self, fontname):
    "change the font for the entire buffer"
    import pango
    font_desc = pango.FontDescription(fontname)
    self.modify_font(font_desc)

    #start, end = self._buffer.get_bounds()
    #if not self._font_tag:
    #  self._font_tag = self._buffer.create_tag(name="font", font=fontname)
    #  #self._font_tag = self._buffer.create_tag(name="font", font="Serif 15")
    #else:
    #  self._font_tag.set_property("font", fontname)
    #  #self._font_tag.set_property("font", "Serif 15")

    #self._buffer.remove_all_tags(start, end)
    #self._buffer.apply_tag(self._font_tag, start, end)

  def toggle_linewrap(self):
    self.linewrap = self.linewrap ^ 1
    if self.linewrap:
      self.set_wrap_mode(gtk.WRAP_WORD)
    else:
      self.set_wrap_mode(gtk.WRAP_NONE)

  def get_modified(self):
    return self._buffer.get_modified()

  def set_modified(self, val):
    return self._buffer.set_modified(val)

  def _claim_selection(self, *args):
    self.have_selection = self.selection_owner_set("PRIMARY")
    # if claiming the selection failed, we ... ignore it
    #if not self._have_selection:
    #  widget.set_active(gtk.FALSE)

  def _selection_handle(self, widget, selection_data, info, time_stamp):
    current_time = self.time.time()
    selection_data.set_text(self._clipboard, len(self._clipboard))

  def _selection_clear(self, widget, event):
    self._have_selection = 0
    return gtk.TRUE

  def cut(self, *args):
    start, end = self._buffer.get_selection_bounds()
    text = self._buffer.get_text(start, end)
    self._clipboard = text
    self._claim_selection()
    self._buffer.delete(start, end)

  def copy(self, *args):
    start, end = self._buffer.get_selection_bounds()
    text = self._buffer.get_text(start, end)
    #self._buffer.copy_clipboard()
    self._clipboard = text
    self._claim_selection()

  def paste(self, *args):
    self._paste_ready = 0
    # request clipboard data
    ret = self.selection_convert("PRIMARY", "STRING")
    start = self.time.time()
    # Wait for the clipboard data to come in
    while not self._paste_ready:
      gtk.mainiteration()
      self.time.sleep(0.001)
      # don't wait more than 1 second for pasted data
      if self.time.time() - start > 1.0:
	return
    self._buffer.insert_at_cursor(self._clipboard)

  def _selection_received(self, widget, selection_data, data):
    # Make sure we got the data in the expected form
    if str(selection_data.type) == "STRING":
	# Print out the string we received
	text = selection_data.get_text()
	self._clipboard = text
	#print "STRING TARGET: %s" % selection_data.get_text()
	self._paste_ready = 1
    else:
	print "Selection was not returned as \"STRING\""
    return gtk.FALSE

class Editor:
  def __init__(self, cfg, filename=None, text="", title=None, mode="normal", transient=None):
    self.cfg = cfg
    
    self.filename = filename
    self.title = title
    self.mode = mode
    
    self.win = gtk.Window()
    win = self.win
    win.connect("delete_event", gtk.mainquit)
    self.set_title(0)
    win.set_default_size(self.cfg.pedit_width,
                         self.cfg.pedit_height)

    if self.mode == "mini":
      win.set_modal(1)
      win.set_position(gtk.WIN_POS_MOUSE)

    if transient:
      win.set_transient_for(transient)


    self.vbox = gtk.VBox()
    win.add(self.vbox)
    
    self._make_menus()
    
    t = gtk.ScrolledWindow()
    t.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
    self.vbox.pack_start(t, 1, 1)
    
    self.editor_widget = EditWidget()
    editor = self.editor_widget
    # make sure we notice when the buffer gets modified
    editor.connect("cut-clipboard", self._modified)
    editor.connect("paste-clipboard", self._modified)
    #editor.connect("delete-from-cursor", self._modified)
    # FIXME: "insert-at-cursor" doesn't work?
    #editor.connect("insert-at-cursor", self._modified)
    # Oh hell, why not be extra thorough?
    editor.connect("key-press-event", self._modified)
    editor.connect("key-release-event", self._modified)
    editor.connect("button-press-event", self._modified)
    editor.connect("button-release-event", self._modified)

    if text:
      editor.set_text(text)
      editor.set_modified(0)
      
    self.fontnum = 0
    editor.set_font(self.cfg.pedit_fonts[self.fontnum])
    t.add(editor)

    win.show_all()
    
  def set_title(self, modified=0):
    if self.title:
      if modified:
	self.win.set_title(self.title + "*")
      else:
	self.win.set_title(self.title)
      return

    title = self.filename
    if not title:
      title = "(new)"
    if modified:
      self.win.set_title("%s: %s*" % (program_name, title))
    else:
      self.win.set_title("%s: %s" % (program_name, title))

  def _make_menus(self):
    self.ag = gtk.AccelGroup()
    self.win.add_accel_group(self.ag)
    self.itemf = gtk.ItemFactory(gtk.MenuBar, "<main>", self.ag)
    # create menu differently depending on our current mode...
    # (don't show "file" items if we shouldn't, etc...)
    menu = []
    if self.mode in ["normal", "mini"]:
      menu.extend([
	("/_File",       None, None, 0, "<Branch>"),
	("/File/_New",    None, self._file_new, 0, ""),
	("/File/_Open",  "<Control>O", self._file_open, 0, ""),
	("/File/_Reload",  "", self._reload, 0, ""),
	("/File/_Save",  "<Control>S", self._save, 0, ""),
	("/File/Save _As...",  "<Alt>A", self._save_as, 0, ""),
	("/File/_Quit",  "<Control>Q", self._quit, 0, ""),
	])
    menu.extend([
      ("/_Edit",       None, None, 0, "<Branch>"),
      ("/Edit/Cu_t",   "<Control>X", self._cut, 0, ""),
      ("/Edit/_Copy",  "<Control>C", self._copy, 0, ""),
      ("/Edit/_Paste", "<Control>V", self._paste, 0, ""),
      ("/_Misc",       None, None, 0, "<Branch>"),
      ("/Misc/Change _Font",   None, self._change_font, 0, ""),
      ("/Misc/Toggle _Linewrap",   None, self._toggle_linewrap, 0, ""),
      #("/_About",       None, None, 0, "<Branch>"),
      #("/About/About %s" % (program_name), None, self.about_screen, 0, ""),
      ])
    if self.mode in ["mini"]:
      menu.extend([
	("/Done",  "<Control>Return", self._quit, 0, ""),
	])

    self.itemf.create_items(menu)
    self.menubar = self.itemf.get_widget("<main>")
    self.vbox.pack_start(self.menubar, fill=0, expand=0)
        
    # Extra keypresses to handle
    #foo = gtk.ItemFactory(gtk.MenuBar, "<keydefs>", self.ag)
    #foo.create_items([
    #    ("/foo",       None, None, 0, "<Branch>"),
    #    ("/foo/play",  "P", self._play_button, 0, ""),
    #    ])

  def _modified(self, *args):
    if self.editor_widget.get_modified():
      self.set_title(1)

  def _cut(self, *args):
    self.editor_widget.cut()

  def _copy(self, *args):
    self.editor_widget.copy()

  def _paste(self, *args):
    self.editor_widget.paste()

  def _toggle_linewrap(self, *args):
    self.editor_widget.toggle_linewrap()

  def _change_font(self, *args):
    self.fontnum = self.fontnum + 1
    if self.fontnum >= len(self.cfg.pedit_fonts):
      self.fontnum = 0
    self.editor_widget.set_font(self.cfg.pedit_fonts[self.fontnum])

  def _file_new(self, *args):
    self.editor_widget.set_text("")
    self.filename = ""
    self.set_title(1)
    
  def _file_open(self, *args):
    filename = self.file_selector("Open...", "")
    if filename:
      self.load(filename)
    
  def _reload(self, *args):
    if self.filename:
      self.load(self.filename)
    
  def _save(self, *args):
    self.save()

  def _save_as(self, *args):
    filename = self.file_selector("Save As...", "")
    if filename:
      self.save(filename)
    
  def file_selector(self, title, filename):
    self.file_selector_cb_button = None
    
    fs = gtk.FileSelection(title)
    if filename:
      fs.set_filename(filename)
    fs.ok_button.connect("clicked", self.file_selector_ok_cb)
    fs.cancel_button.connect("clicked", self.file_selector_cancel_cb)
    fs.show_all()
    gtk.mainloop()

    if self.file_selector_cb_button == "ok":
      filename = fs.get_filename()
      fs.hide()
      return filename
    else:
      fs.hide()
      return None

  def file_selector_ok_cb(self, *args):
    #print "ok"
    self.file_selector_cb_button = "ok"
    gtk.mainquit()

  def file_selector_cancel_cb(self, *args):
    #print "cancel"
    self.file_selector_cb_button = "cancel"
    gtk.mainquit()

  def load(self, filename):
    print "Loading... %s" % (filename)
    import string
    import os
    d = os.path.dirname(filename)
    f = os.path.basename(filename)
    self.filename = f
    self.set_title(0)
    try:
      if d:
	os.chdir(d)
      file = open(f, "r")
      lines = file.readlines()
      text = string.join(lines, "")
      self.editor_widget.set_text(text)
      self.editor_widget.set_font(self.cfg.pedit_fonts[self.fontnum])
      file.close()
      self.editor_widget.set_modified(0)
      self.set_title(0)
    except IOError, msg:
      print "Cannot open file %s: %s" % (filename, msg)

  def save(self, filename=None):
    if filename:
      import os
      d = os.path.dirname(filename)
      f = os.path.basename(filename)
      if d:
	try:
	  os.chdir(d)
	except:
	  pass
      self.filename = f
      self.set_title(0)
    if not self.filename:
      self._save_as()
      return
    if self.cfg.pedit_make_backups:
      import shutil #os

      try:
        #os.rename(self.filename, "%s~" % (self.filename))
        shutil.copy(self.filename, "%s~" % (self.filename))
      except Exception, msg:
	print "Unable to create backup file (%s~): %s" % (self.filename, msg)
        pass

    try:
      file = open(self.filename, "w+")
      file.write(self.editor_widget.get_text())
      file.close()
      self.editor_widget.set_modified(0)
      self.set_title(0)
    except Exception, msg:
      ButtonChoiceWindow()("Error Saving",
                         "Could not save %s: %s" % (self.filename, msg),
                         ["Ok"])

  def get_text(self):
    return self.editor_widget.get_text()

  def run(self):
    gtk.mainloop()

  def _quit(self, *args):
    if self.mode == "mini":
      self.quit()
      return

    if self.editor_widget.get_modified():
      ans = ButtonChoiceWindow()("Unsaved", "File is unsaved.  Quit anyway?",
                                 [("Cancel", "no"), ("Quit", "quit")], 0, 0,
				 self.win)
      if ans == "quit":
	self.quit()
    else:
      self.quit()

  def quit(self):
    self.win.hide()
    self.win.unrealize()
    gtk.mainquit()

def run(args):
  import pycfg
  cfg = pycfg.config(program_name)
  cfg.load()

  edit = Editor(cfg)
  if len(args) > 0:
    edit.load(args[0])
    
  edit.run()

if __name__ == "__main__":
  import sys
  run(sys.argv[1:])
   
